# Problem structures

import sys

from errors import *
import common



class TimePoint(object):

    def __init__(self, name, controllable):
        self.name = name
        self.controllable = controllable
        return

    pass



class AtomicConstraint(object):

    def __init__(self, source, dest, lower, upper, contingent=None):

        self.contingent = contingent

        if self.checkValue(lower, False) and self.checkValue(upper, True):
            self.source = source
            self.dest = dest
            self.lowerBound = lower
            self.upperBound = upper
        else:
            raise MalformationError()

        return


    def checkValue(self, v, upper):
        if isinstance(v, str):
            return (upper and (v == common.INFINITY)) or ((not upper) and (v == common.MINUS_INFINITY))
        else:
            return isinstance(v, int) or isinstance(v, float)

        assert False
        return False

    def __str__(self):
        return ("(%s - %s) in [%s, %s]" % (self.dest.name, self.source.name, str(self.lowerBound), str(self.upperBound)))

    def __repr__(self):
        return str(self)


    def __eq__(self, oth):
        return str(self) == str(oth)

    def __hash__(self):
        return hash(str(self))

    pass



class Constraint(object):

    def __init__(self, atoms, contingent=False, fixBin=False):
        self.atoms = list(atoms)
        self.contingent = contingent

        if fixBin and len(atoms) > 1:
            self.fixBinary()

        return


    def isSimple(self):
        return len(self.atoms) == 1


    def hasUncontrollables(self):
        for a in self.atoms:
            if (not a.source.controllable) or \
                    (not a.dest.controllable):
                return True
        return False


    def isBinary(self):
        s = None
        d = None

        for a in self.atoms:
            if s is None and d is None:
                s = a.source
                d = a.dest
                assert s is not None
                assert d is not None
            else:
                if a.source != s or a.dest != d:
                    return False

        return True


    def fixBinary(self):
        s = None
        d = None

        for a in self.atoms:
            if s is None and d is None:
                s = a.source
                d = a.dest
                assert s is not None
                assert d is not None
            else:
                if a.source == d and a.dest == s:
                    tmp = a.source
                    a.source = a.dest
                    a.dest = tmp

                    l = a.lowerBound
                    u = a.upperBound

                    if l != 0 and u != 0:
                        if l == common.MINUS_INFINITY:
                            a.upperBound = common.INFINITY
                        else:
                            a.upperBound = -1 * l

                            if u == common.INFINITY:
                                a.lowerBound = common.MINUS_INFINITY
                            else:
                                a.lowerBound = -1 * u
        return


    def __str__(self):
        res = ""
        for a in self.atoms:
            res += str(a)
            if a != self.atoms[-1]:
                res += " v "
        return res


    def __repr__(self):
        return str(self)

    pass



class DisjunctivityLevel(object):

    (NONE, BINARY, FULL) = range(0,3)

    pass



class Problem(object):

    def __init__(self):
        self.timePoints = {}
        self.constraints = []

        self.positive = True
        self.disjunctivity = DisjunctivityLevel.NONE
        self.game = False
        return


    def createTimePoint(self, name, controllable):
        if not controllable:
            self.game = True

        if type(name) is int:
            name = ("X%d" % name)

        if name in self.timePoints:
            raise RedefinitionError()

        n = TimePoint(name, controllable)
        self.timePoints[name] = n
        return n


    def getTimePoint(self, name):
        if type(name) is int:
            name = ("X%d" % name)

        if name in self.timePoints:
            return self.timePoints[name]
        raise NoSuchElementError()


    def createConstraint(self, atoms, contingent=False, fixBinary=False):
        assert len(atoms) > 0;

        c = Constraint(atoms, contingent, fixBinary)
        self.constraints.append(c)

        assert (not contingent) or c.isBinary()

        if len(atoms) > 1:
            # at least binary
            if self.disjunctivity != DisjunctivityLevel.FULL:
                if c.isBinary():
                    self.disjunctivity = DisjunctivityLevel.BINARY
                else:
                    self.disjunctivity = DisjunctivityLevel.FULL

        for a in atoms:
            if a.lowerBound == common.MINUS_INFINITY or a.lowerBound < 0:
                self.positive = False
                break

            if a.upperBound != common.INFINITY and a.upperBound < 0:
                self.positive = False
                break

        return c


    def getDensity(self):
        e = len(self.constraints)
        v = len(self.timePoints)
        return (float(2 * e) / float(v * (v - 1)))


    def toFile(self, stream=sys.stdout):
        '''Print in AMI format'''

        stream.write("%d %d\n" % (len(self.timePoints), len(self.constraints)))

        for n in self.timePoints.values():
            stream.write("%s " % n.name)
            if n.controllable:
                stream.write("c\n")
            else:
                stream.write("u\n")

        for c in self.constraints:
            cons = "f"
            if c.contingent:
                cons = "c"

            stream.write("%d %s " % (len(c.atoms), cons))

            for a in c.atoms:
                s = a.source.name
                d = a.dest.name
                l = str(a.lowerBound)
                u = str(a.upperBound)

                stream.write("%s %s %s %s " % (s, d, l, u))

            stream.write("\n")
        return


    def fromFile(self, filename):
        f = open(filename, "rt")
        entries = f.read().split()

        nn = int(entries[0])
        ne = int(entries[1])

        pointer = 2
        for i in range(0, nn):
            name = entries[pointer]
            cflag = entries[pointer+1]

            self.createTimePoint(name, (cflag=="c"))
            pointer = pointer + 2

        for i in range(0, ne):
            ndisjunction = int(entries[pointer])

            cflag = entries[pointer+1]
            contingent = (cflag=="c")

            pointer = pointer + 2

            atoms = set()
            for _ in range(0, ndisjunction):
                s = self.getTimePoint(entries[pointer])
                d = self.getTimePoint(entries[pointer+1])

                l = entries[pointer+2]
                u = entries[pointer+3]

                try:
                    if u != "+inf" and u != "-inf":
                        u = float(u)

                    if l != "+inf" and l != "-inf":
                        l = float(l)
                except Exception:
                    raise ParsingError("Expected intger or infinity keywords", nn + i + 1)

                assert (not contingent) or (s.controllable and (not d.controllable))

                at = None
                try:
                    at = AtomicConstraint(s, d, l,u, contingent)
                except MalformationError:
                    pass

                if at is not None:
                    atoms.add(at)

                pointer = pointer + 4

            if len(atoms) == 0:
                raise TriviallyInconsistentError("All the constraints in line %d are trivially false")
            self.createConstraint(atoms, contingent)
        return


    def printDot(self, stream=sys.stdout):
        '''Prints the given network in DOT format'''

        if self.disjunctivity == DisjunctivityLevel.FULL:
            raise UnsupportedError()


        stream.write("digraph {\n");

        for v in self.timePoints.values():
            cs = ""
            if not v.controllable:
                cs = "double"

            stream.write("node [shape=%scircle] \"%s\";\n" % (cs, v.name))

        stream.write("\n")


        for c in self.constraints:
            s = c.atoms[0].source.name
            d = c.atoms[0].dest.name

            style = "dashed"
            atype = "empty"
            if c.contingent:
                style = "solid"
                atype = "normal"

            stream.write(("\"%s\" -> \"%s\" [ arrowhead=\"%s\", style=\"%s\", label = \"" %
                          (s, d, atype, style)))


            count = 0
            for a in c.atoms:
                l = a.lowerBound
                u = a.upperBound

                stream.write("[")
                stream.write(str(l))
                stream.write(", ")
                stream.write(str(u))
                stream.write("]")

                count = count + 1
                if count < len(c.atoms):
                    stream.write(" | ")

            stream.write("\"];\n")


        stream.write("\n")
        stream.write("}\n");
        return

    pass



class Interval(object):

    def __init__(self, name, l, u, contingent):
        if type(name) is int:
            name = ("X%d" % name)

        self.name = name
        self.lowerBound = l
        self.upperBound = u
        self.contingent = contingent
        return

    pass



class AllenRelation(object):

    (BEFORE,AFTER,STARTS,STARTED_BY,FINISHES,FINISHED_BY,MEETS,MET_BY,DURING,CONTAINED,OVERLAPS,OVERLAPPED_BY,EQUAL, S_E, E_S, S_S, E_E) = range(0,17)

    def __init__(self, source, dest, typeid, bounds):
        self.source = source
        self.dest = dest
        self.typeid = typeid
        self.bounds = bounds
        return

    pass


class AllenConstraint(object):

    def __init__(self, relations):
        self.relations = list(relations)
        return

    def isBinary(self):
        s = None
        d = None

        for a in self.relations:
            if s is None and d is None:
                s = a.source
                d = a.dest
                assert s is not None
                assert d is not None
            else:
                if a.source != s or a.dest != d:
                    return False
        return True

    pass



class IntervalProblem(object):

    def __init__(self):
        self.intervals = {}
        self.constraints = []
        return


    def createInterval(self, name, l, u, contingent):
        if name in self.intervals:
            raise RedefinitionError()
        else:
            n = Interval(name, l, u, contingent)
            self.intervals[name] = n
            return n


    def getInterval(self, name):
        if name in self.intervals:
            return self.intervals[name]
        else:
            raise NoSuchElementError()


    def createConstraint(self,relations):
        c = AllenConstraint(relations)
        self.constraints.append(c)
        return c


    def getDensity(self):
        e = len(self.constraints)
        v = len(self.intervals)
        return (float(2 * e) / float(v * (v - 1)))

    pass



class IntervalConverter(object):

    def __init__(self, intervalProblem, fixBinary=False):
        self.problem = intervalProblem

        self.interval2start = {}
        self.interval2end = {}

        self.fixBinary = fixBinary
        return


    def getProblem(self):
        res = Problem()

        # create time points
        for i in self.problem.intervals.values():
            s = res.createTimePoint(i.name + "_start", True)
            e = res.createTimePoint(i.name + "_end", not i.contingent)

            self.interval2start[i] = s
            self.interval2end[i] = e

            res.createConstraint([AtomicConstraint(s, e, i.lowerBound, i.upperBound,i.contingent)], i.contingent)


        for c in self.problem.constraints:
            disj = []
            for r in c.relations:
                disj.append(self.convertRelation(r))


            cnf = []
            for d in disj:
                if cnf == []:
                    for a in d:
                        cnf.append([a])
                else:
                    tmp = []
                    for a in d:
                        for c in cnf:
                            nc = list(c)
                            nc.append(a)
                            tmp.append(nc)
                    cnf = tmp

            for c in cnf:
                res.createConstraint(c, False, self.fixBinary)


        return res


    def convertRelation(self, r):
        # res = []
        # if r.typeid == AllenRelation.BEFORE:
        #     res.append(AtomicConstraint(self.interval2end[r.source], self.interval2start[r.dest], r.bounds[0], r.bounds[1]))
        # elif r.typeid == AllenRelation.AFTER:
        #     res.append(AtomicConstraint(self.interval2end[r.dest], self.interval2start[r.source], r.bounds[0], r.bounds[1]))
        # elif r.typeid == AllenRelation.STARTS:
        #     res.append(AtomicConstraint(self.interval2start[r.source], self.interval2start[r.dest], 0, 0))
        # elif r.typeid == AllenRelation.STARTED_BY:
        #     res.append(AtomicConstraint(self.interval2start[r.dest], self.interval2start[r.source], 0, 0))
        # elif r.typeid == AllenRelation.FINISHES:
        #     res.append(AtomicConstraint(self.interval2end[r.source], self.interval2end[r.dest], 0, 0))
        # elif r.typeid == AllenRelation.FINISHED_BY:
        #     res.append(AtomicConstraint(self.interval2end[r.dest], self.interval2end[r.source], 0, 0))
        # elif r.typeid == AllenRelation.MEETS:
        #     res.append(AtomicConstraint(self.interval2end[r.source], self.interval2start[r.dest], 0, 0))
        # elif r.typeid == AllenRelation.MET_BY:
        #     res.append(AtomicConstraint(self.interval2end[r.dest], self.interval2start[r.source], 0, 0))
        # elif r.typeid == AllenRelation.DURING:
        #     res.append(AtomicConstraint(self.interval2start[r.dest], self.interval2start[r.source], r.bounds[0], r.bounds[1]))
        #     res.append(AtomicConstraint(self.interval2end[r.source], self.interval2end[r.dest], r.bounds[2], r.bounds[3]))
        # elif r.typeid == AllenRelation.CONTAINED:
        #     res.append(AtomicConstraint(self.interval2start[r.source], self.interval2start[r.dest], r.bounds[0], r.bounds[1]))
        #     res.append(AtomicConstraint(self.interval2end[r.dest], self.interval2end[r.source], r.bounds[2], r.bounds[3]))
        # elif r.typeid == AllenRelation.OVERLAPS:
        #     res.append(AtomicConstraint(self.interval2start[r.dest], self.interval2end[r.source], r.bounds[0], r.bounds[1]))
        #     res.append(AtomicConstraint(self.interval2end[r.source], self.interval2end[r.dest], 0, common.INFINITY))
        # elif r.typeid == AllenRelation.OVERLAPPED_BY:
        #     res.append(AtomicConstraint(self.interval2start[r.source], self.interval2end[r.dest], r.bounds[0], r.bounds[1]))
        #     res.append(AtomicConstraint(self.interval2end[r.dest], self.interval2end[r.source], 0, common.INFINITY))
        # elif r.typeid == AllenRelation.EQUAL:
        #     res.append(AtomicConstraint(self.interval2start[r.source], self.interval2start[r.dest], 0, 0))
        #     res.append(AtomicConstraint(self.interval2end[r.source], self.interval2end[r.dest], 0, 0))
        # else:
        #     assert False, ("Typeod cannot be  %d" % r.typeid)


        res = []
        if r.typeid == AllenRelation.BEFORE:
            res.append(AtomicConstraint(self.interval2end[r.source], self.interval2start[r.dest], 0, common.INFINITY))
        elif r.typeid == AllenRelation.AFTER:
            res.append(AtomicConstraint(self.interval2end[r.dest], self.interval2start[r.source], 0, common.INFINITY))
        elif r.typeid == AllenRelation.STARTS:
            res.append(AtomicConstraint(self.interval2start[r.source], self.interval2start[r.dest], 0, 0))
        elif r.typeid == AllenRelation.STARTED_BY:
            res.append(AtomicConstraint(self.interval2start[r.dest], self.interval2start[r.source], 0, 0))
        elif r.typeid == AllenRelation.FINISHES:
            res.append(AtomicConstraint(self.interval2end[r.source], self.interval2end[r.dest], 0, 0))
        elif r.typeid == AllenRelation.FINISHED_BY:
            res.append(AtomicConstraint(self.interval2end[r.dest], self.interval2end[r.source], 0, 0))
        elif r.typeid == AllenRelation.MEETS:
            res.append(AtomicConstraint(self.interval2end[r.source], self.interval2start[r.dest], 0, 0))
        elif r.typeid == AllenRelation.MET_BY:
            res.append(AtomicConstraint(self.interval2end[r.dest], self.interval2start[r.source], 0, 0))
        elif r.typeid == AllenRelation.DURING:
            res.append(AtomicConstraint(self.interval2start[r.dest], self.interval2start[r.source], 0, common.INFINITY))
            res.append(AtomicConstraint(self.interval2end[r.source], self.interval2end[r.dest], 0, common.INFINITY))
        elif r.typeid == AllenRelation.CONTAINED:
            res.append(AtomicConstraint(self.interval2start[r.source], self.interval2start[r.dest], 0, common.INFINITY))
            res.append(AtomicConstraint(self.interval2end[r.dest], self.interval2end[r.source], 0, common.INFINITY))
        elif r.typeid == AllenRelation.OVERLAPS:
            res.append(AtomicConstraint(self.interval2start[r.dest], self.interval2end[r.source], 0, common.INFINITY))
            res.append(AtomicConstraint(self.interval2end[r.source], self.interval2end[r.dest], 0, common.INFINITY))
        elif r.typeid == AllenRelation.OVERLAPPED_BY:
            res.append(AtomicConstraint(self.interval2start[r.source], self.interval2end[r.dest], 0, common.INFINITY))
            res.append(AtomicConstraint(self.interval2end[r.dest], self.interval2end[r.source], 0, common.INFINITY))
        elif r.typeid == AllenRelation.EQUAL:
            res.append(AtomicConstraint(self.interval2start[r.source], self.interval2start[r.dest], 0, 0))
            res.append(AtomicConstraint(self.interval2end[r.source], self.interval2end[r.dest], 0, 0))


        # Magic relations
        elif r.typeid == AllenRelation.S_S:
            res.append(AtomicConstraint(self.interval2start[r.source], self.interval2start[r.dest],  r.bounds[0], r.bounds[1]))
        elif r.typeid == AllenRelation.S_E:
            res.append(AtomicConstraint(self.interval2start[r.source], self.interval2end[r.dest],  r.bounds[0], r.bounds[1]))
        elif r.typeid == AllenRelation.E_S:
            res.append(AtomicConstraint(self.interval2end[r.source], self.interval2start[r.dest],  r.bounds[0], r.bounds[1]))
        elif r.typeid == AllenRelation.E_E:
            res.append(AtomicConstraint(self.interval2end[r.source], self.interval2end[r.dest],  r.bounds[0], r.bounds[1]))

        else:
            assert False, ("Typeid cannot be  %d" % r.typeid)

        return res
